"""
Tools to implement the command line interface
=============================================

This module is used internally to implement the command line interface.
In addition, it can be used to run the commands that are available
in the command line interface without the need to leave python:

```python
from bussilab.cli import cli
cli("-h")
cli("wham -b bias") # provide the command line as a string
cli(["wham", "-b", "bias"]) # alternatively use a list
```

The documentation of all the commands can be found in the
[cli_documentation](cli_documentation.html) submodule.

Notice however that these commands typically have alternative python
implementations that allow you to work directly on data structures
and are thus more flexible.
For instance, `bussilab.wham.wham(bias)`, where `bias` is a numpy array,
is often more convenient than `bussilab.cli.cli("wham -b bias")`,
where `bias` is a file.

"""

# Minimal imports here.

# Support argument parsing:
import argparse

# Support static type checks:
from typing import Optional, Union, List, Callable, Dict, Tuple

# Do not import heavy modules here since they could slow down the command line interface
# It's fine instead to load them lazily in functions where they are needed.

class _Argument():
    """Special class returned by the decorators present in this module."""
    def __init__(self, func: Union[Callable, "_Argument"]):
        if isinstance(func, _Argument):
            self.func: Callable = func.func
            self.calls: List[Callable] = func.calls
        else:
            self.func = func
            self.calls = []
    def __call__(self, *args, **kwargs):
        msg = ("Functions that are registered for the command line are not directly callable.\n"
               "Use cli() instead.")
        raise NotImplementedError(msg)

# list of registered commands:
_commands: List[Tuple[str, str, str, _Argument, Dict]] = []


class _ExtendedParser():
    """Internal class used to attach subparsers help to parser."""
    def __init__(self,
                 parser: argparse.ArgumentParser,
                 subparsers_help: Dict[str, str],
                 subparsers_doc: Dict[str, str]
                ):
        self.parser = parser
        self.subparsers_help = subparsers_help
        self.subparsers_doc = subparsers_doc

def _create_eparser(prog: Optional[str] = None) -> _ExtendedParser:
    """Create a `parser` object."""
    parser = argparse.ArgumentParser(prog=prog, description="command line tool")
    parser.add_argument("--version", action='store_true', dest="_version")
    subparsers = parser.add_subparsers(title="Subcommands")
    subparsers_help = {}
    subparsers_doc = {}
    for a in _commands:
        (name, h, description, func, kwargs) = a
        p = [subparsers.add_parser(name, help=h, description=description, **kwargs)]
        f = p[0]
        for call in reversed(func.calls):
            p = call(p)
        if len(p) > 1:
            msg = "Non matching group/endgroup"
            raise TypeError(msg)
        p[0].set_defaults(_func=func.func)
        subparsers_help[name] = f.format_help()
        if func.func.__doc__ is not None:
            subparsers_doc[name] = func.func.__doc__
        else:
            subparsers_doc[name] = ""
    return _ExtendedParser(parser, subparsers_help, subparsers_doc)

def command(name: str, help: Optional[str] = None, description: Optional[str] = None, **kwargs):
    """Decorator that registers a function as a subcommand.

       This decorator should be written **before** the other decorators
       `bussilab.cli.arg`, `bussilab.cli.group`, and `bussilab.cli.endgroup`.

       Parameters
       ----------
       name : str
           Name of the subcommand (will be used on the command line)

       help : str
           Short help message for the subcommand (one line).

       description : str, optional
           Longer description. If not provided, it is set to a copy of `help`.

       kwargs
           Other parameters are passed as is to the `add_parser` function of `argparse`.

       Examples
       --------

       Simple command line tool that accepts a single `--out` argument followed by a string
       and call the function `do_something` with that string as an argument.

       ```python
       from bussilab.cli import command, arg

       @command("subcommand")
       @arg("--out")
       def myfunc(out):
           do_something(out)
       ```
    """
    if description is None:
        description = help
    def add_command(func):
        func = _Argument(func)
        _commands.append((name, help, description, func, kwargs))
        return func
    return add_command

def arg(*name, **kwargs):
    """ Decorator that adds an argument to a command line tool.

        Parameters are passed to the `parser.add_argument()` function.
        It should be written **after** the `bussilab.cli.command` decorator.
    """
    def call(p):
        p[-1].add_argument(*name, **kwargs)
        return p
    def add_argument(func):
        func = _Argument(func)
        func.calls.append(call)
        return func
    return add_argument

def group(title: Union[str, Callable, None] = None,
          description: Optional[str] = None,
          exclusive: Optional[bool] = None,
          required: Optional[bool] = None):
    """ Decorator that adds a group of arguments for a command line tool.

        It should be written **after** the `bussilab.cli.command` decorator.
        It should be followed by a number of `bussilab.cli.arg` decorators and by a closing
        `bussilab.cli.endgroup` decorator.

        Parameters
        ----------
        title : str
            The name of the group. Can only be used for non exclusive groups.

        description : str
            A description of the group. Can only be used for non exclusive groups.

        exclusive : bool
            If True, the arguments belonging to this group are mutually exclusive

        required : bool
            If True, one of the arguments at least should be passed.
            Can only be used for exclusive groups.

        Examples
        --------

        This is a simple command line tool that accepts three arguments (`-a`, `-b`, or `-c`),
        mutually exclusive. When ran, it will just print booleans showing if these arguments
        were passed.

        ```python
        from bussilab.cli import command, group, arg, endgroup

        @command("doit")
        @group(exclusive=True)
        @arg("-a", action='store_true')
        @arg("-b", action='store_true')
        @arg("-c", action='store_true')
        @endgroup
        def check(a, b, c):
           print(a, b, c)
        ```
    """
    noarg = None
    # note: _Argument is callable as well
    if description is None and exclusive is None and required is None and callable(title):
        noarg = title
        title = None
    if exclusive is None:
        exclusive = False
    if not exclusive and required is not None:
        msg = "required can only be used for exclusive groups"
        raise TypeError(msg)
    if exclusive and title is not None:
        msg = "title can only be used for non exclusive groups"
        raise TypeError(msg)
    if exclusive and description is not None:
        msg = "description can only be used for non exclusive groups"
        raise TypeError(msg)
    if exclusive and required is None:
        required = False # default for add_mutually_exclusive_group
    def call(p):
        if exclusive:
            p.append(p[-1].add_mutually_exclusive_group(required=required))
        else:
            p.append(p[-1].add_argument_group(title=title, description=description))
        return p
    def begin_group(func: Union[Callable, _Argument]):
        func = _Argument(func)
        func.calls.append(call)
        return func
    if noarg:
        return begin_group(noarg)
    return begin_group

def endgroup(f: Optional[Callable] = None):
    """ Decorator that ends a group of arguments for a command line tool.

        See `bussilab.cli.group`.
    """
    def call(p):
        if len(p) < 2:
            msg = "Non matching group/endgroup"
            raise TypeError(msg)
        return p[:-1]
    def end_group(func):
        func = _Argument(func)
        func.calls.append(call)
        return func
    if f:
        return end_group(f)
    return end_group

def cli(arguments: Union[str, List[str]] = "",
        *,
        prog: Optional[str] = "", # None is used to imply sys.argv[0]
        use_argcomplete: bool = False,
        throws_on_parser_errors: bool = True
       ) -> Optional[int]:
    """Executes a command line tool from python.

       This is the main function of this module and allows to launch all the subcommands available
       in the command line interface directly from python.

       Parameters
       ----------
       arguments : str or list
           Command line arguments. If a string is passed, it is first split using
           spaces as separators.

       prog : str
           Name of the calling program. It is used to build help texts. **Mostly for internal use**.

       use_argcomplete : bool
           If True, the `autocomplete` function of `argcomplete` module is called on the parser,
           so as to allow autocompletion in the command line tool.
           If `argcomplete` module is not installed, nothing is done and no failure is reported.
           **Mostly for internal use**.

       throws_on_parser_errors : bool
           If True, in case of command line error it throws a TypeError
           exception. **Mostly for internal use**.

       Returns
       -------
       None or int
           If an error happens while parsing, it throws a TypeError exception,
           unless throws_on_parser_errors is set to false, in which ase it
           returns the corresponding error code.
           If an error happens while executing the requested command, an exception is thrown.
           If everything goes well, it returns None.
    """

    # allow passing a single string
    if isinstance(arguments, str):
        arguments = arguments.split()

    func = None

    if prog == "":
        eparser = _eparser
    else:
        eparser = _create_eparser(prog)

    if use_argcomplete:
        # optional feature:
        try:
            import argcomplete
            argcomplete.autocomplete(eparser.parser)
        except ImportError:
            pass

    # Parse options
    # In order to avoid python to crash when cli() is called from python
    # with wrong arguments, it is necessary to intercept the exception.
    try:
        args = vars(eparser.parser.parse_args(arguments))
    except SystemExit as e:
        if e.code != 0:
            if throws_on_parser_errors:
                raise TypeError("Error parsing command line arguments," +
                                " return code is " + str(e.code))
            else:
                return e.code
        return None

    if '_func' in args:
        func = args["_func"]
        del args["_func"]

    main_args = {}
    for i in args.keys():
        if i[0] == "_":
            main_args[i[1:]] = args[i]

    for i in main_args:
        del args["_"+i]

    if "version" in main_args and main_args["version"]:
        from . import __version__
        print(__version__)
        return None

    remove = []
    for i in args.keys():
        if args[i] is None:
            remove.append(i)

    for i in remove:
        del args[i]

    if func:
        func(**args)
    else:
        eparser.parser.print_usage()

    return None

def _main_(prog: Optional[str] = None):

    import sys

    # process command line arguments:
    # prog==None implies argv[0]
    sys.exit(cli(prog=prog, arguments=sys.argv[1:], use_argcomplete=True))

################################################################################

# BELOW ARE ALL THE COMMAND LINE TOOLS

# These functions are made not available from python by their decorators.
# Only to be used from cli() or from the command line.

@command("list", help="List available python modules.")
def _list_submodules():
    from . import list_submodules, describe_submodule
    print("List of available python modules:")
    print()
    l = list_submodules()
    c = max(map(len, l))
    for m in list_submodules():
        print(m, " "*(c-len(m)) + ": " + describe_submodule(m))

@command("check",
         help="Check installed features",
         description="Check installed features")
@group(exclusive=True)
@arg("--import", action='store_true',
     dest="check_import", help="check if all the submodules can be imported")
@endgroup
def _check(check_import: bool = False):
    if check_import:
        from . import import_submodules
        print("Importing all submodules:")
        import_submodules()
        print("All submodules were imported successfully, dependencies seem to be in place")
    else:
        print("nothing to check")

@command("wham", help="Perform binless WHAM",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("-b", "--bias", help="File containing bias potential", required=True)
@arg("-o", "--out", help="Output file with weights")
@arg("--use-frame-weight", action="store_true")
@arg("--traj-weight", nargs="*", type=float)
@arg("-T", "--temperature", type=float, default=1.0,
     help="system temperature in energy units", dest="T")
@arg("-m", "--maxiter", type=int, default=1000, help="maximum number of iterations")
@arg("-t", "--threshold", type=int, default=1e-40, help="threshold for convergence")
@arg("-v", "--verbose", action="store_true")
def _wham(**args):
    from . import wham
    import numpy as np
    # load bias file:
    args["bias"] = np.loadtxt(args["bias"])

    # with --use-frame-weight, last column in bias is interpreted as frame weight
    if "use_frame_weight" in args:
        if args["use_frame_weight"]:
            args["frame_weight"] = args["bias"][:, -1]
            args["bias"] = args["bias"][:, :-1]
        del args["use_frame_weight"]

    if "out" in args:
        outfile = args["out"]
        del args["out"]
    else:
        outfile = None

    ret = wham.wham(**args)

    print("# nit:", ret.nit)
    print("# eps:", ret.eps)
    print("# logZ:", np.array_str(ret.logZ, max_line_width=np.inf))

    if outfile is None:
        for i in ret.logW:
            print(i)
    else:
        with open(outfile, "w") as f:
            for i in ret.logW:
                print("%8.4f"%i, file=f)

@command("jrun", help="Run jupyter server",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("-d", "--dry-run", action='store_true',
     help="show command instead of executing it")
@arg("--port", help="set port", default=0)
@arg("--screen-cmd", help="screen command", default="screen")
@arg("--python-exec", help="python executable", default="")
@arg("-S", "--sockname", help="screen sockname", default="jupyter-server")
@arg("--no-screen", help="do not run screen", action="store_true")
@arg("--detach", help="detach screen", action="store_true")
def _jrun(**kargs):
    """
       This is a tool to run a jupyter server within a screen command.
       The typical usage would be
       ```bash
       cd /path/to/your/notebook/dir
       bussilab jrun
       ```
       A free port is identified first (can be overridden with the `--port`
       option) and a jupyter server is then run  inside a `screen` instance.
       You will thus have to type `CTRL+aCTRL+d` in order to detach the screen
       letting it run in the background.
       To avoid this you can use the `--no-screen` flag. In this case
       however make sure you do not close the terminal where the server is
       running.
    """
    from . import jremote
    jremote.run_server(**kargs)

@command("jremote", help="Run jupyter client",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("server", help="server URL (e.g. giorgione.phys.sissa.it)")
@arg("-d", "--dry-run", action='store_true',
     help="show command instead of executing it")
@arg("-l", "--list-only", action='store_true',
     help="only report a list or servers")
@arg("--port", help="set port", default=0)
@arg("-i", "--index", type=int, help="choose server, by default interactive choice", default=0)
@arg("--python-exec",
     help="remote python executable (e.g. module load python3 python-home; python)",
     default="python")
@arg("--open-cmd", help="open command (detected automatically by default)", default="")
def _jremote(**kargs):
    """
       This is a tool to connect to a remote running jupyter server.
       The typical usage would be
       ```bash
       bussilab jremote server.url
       ```

       A list of jupyter servers running on the selected machines will be
       shown, and one of them can be picked typing its progressive number.
       In case there is a single server running, it will be opened by default.

       Notice that if the name of the python executable on the server is
       different from plain `python` you can override it with
       ``--python-exec``. You can also run other scripts before, for instance
       loading relevant modules:
       ```bash
       bussilab jremote giorgione.phys.sissa.it --python-exec \
           --python-exec "module load python3 python-home; python3"
       ```

       If you recurrently connect to the same workstation, it is convenient
       to write a small script like this one, call it `jremote` and put it in
       your path:
       ```bash
       export PYTHONPATH=/path/to/bussilab/source
       python -m bussilab jremote giorgione.phys.sissa.it --python-exec "module load python3 python-home ; python3"
       ```
       Here replace `python` with the name of your python interpreter (might be
       `python3.7`).

    """
    from . import jremote
    jremote.remote(**kargs)

@command("pip_upgrade_all", help="Upgrade all packages with pip",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("--user", action='store_true',
     help="install/upgrade in user location")
def _pip_upgrade_all(**kargs):
    """
       This is a tool to upgrade all your packages with pip.
       It is a convenient way to upgrade all the packages without
       the need to list them explicitly.

       Warning: this uses pip, so it might not work as expected
       if you are working in conda.

       The typical usage would be
       ```bash
       bussilab pip_upgrade_all
       ```
       If you installed packages in your home you should use
       ```bash
       bussilab pip_upgrade_all --user
       ```
    """
    from . import pip
    pip.upgrade_all(**kargs)

@command("notify", help="Send a notification to Slack",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("-m", "--message", help="message")
@group(exclusive=True)
@arg("-c","--channel", help="channel (check ~/.bussilabrc by default)")
@arg("-u","--update", help="url of the message to be updated")
@arg("-d","--delete", help="url of the message to be deleted")
@endgroup
@arg("-t","--title", help="url of the message to be deleted")
@arg("--type", help="'plain_text' or 'mrkdwn'", default='mrkdwn')
@arg("--token", help="token (check ~/.bussilabrc by default")
def _notify(**kargs):
    """
       This is a tool to send a notification to Slack.
       See the documentation of `bussilab.notify`.
    """
    from . import notify
    msg=notify.notify(**kargs)
    print(msg)

@command("cron", help="Run cron",
          formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@arg("--quick-start", help="run immediately", action="store_true")
@arg("--cron-file", help="path to cron file")
@arg("--screen-cmd", help="screen command", default="screen")
@arg("--no-screen", help="do not run screen", action="store_true")
@arg("-S", "--sockname", help="screen sockname", default="cron")
@arg("--python-exec", help="python executable", default="")
@arg("--detach", help="detach screen", action="store_true")
@arg("--period", help="period (seconds)", default=3600, type=int)
@arg("--max-times", help="maximum number of calls", default=None, type=int)
def _cron(**kargs):
    from . import cron
    cron.cron(**kargs)

@command("required", help="print requirements",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
@group(exclusive=True)
@arg("--macports", help="conda syntax", action="store_true")
@arg("--conda", help="macports syntax", action="store_true")
@endgroup
@arg("--pyver", help="pyversion (e.g. 38), for macports only", default="")
def _required(macports: bool = False, conda: bool = False, pyver: str = ""):
    if conda:
        from . import required_conda
        print(required_conda())
    elif macports:
        from . import required_macports
        print(required_macports(pyver))
    else:
        from . import required_pip
        print(required_pip())


# Here we create a general eparser to be used from cli with prog=""
_eparser = _create_eparser("")



"""
Module with artificial neural networks.

ANN can be constructed with `cuda=True`, in which case it will use cudamat.
"""

import numpy as np

try:
    import cudamat as cm # pylint: disable=import-error
    _HAS_CUDAMAT=True
except ModuleNotFoundError:
    _HAS_CUDAMAT=False

from .coretools import ensure_np_array

def _softplus(x):
    return np.log(1 + np.exp(-np.abs(x))) + np.maximum(x,0)

def _sigmoid(x):
    return np.exp(-np.logaddexp(0, -x))

def _relu(x):
    return np.maximum(x,0)

def _drelu(x):
    return np.heaviside(x,0.5)

def _sigmoid_cudamat(x):
    cm.sigmoid(x)

def _relu_cudamat(x):
    x.maximum(0.0)

def _drelu_cudamat(x):
    x.greater_than(0.0)

class ANN:
    # constructor, allocate space for parameters
    def __init__(self,layers,random_weights=False,init_b=0.0,activation="softplus",cuda=False):

        if cuda is None:
            cuda = _HAS_CUDAMAT

        self.cuda=cuda

        if not self.cuda:
            if activation == 'softplus':
                self._activation=_softplus
                self._dactivation=_sigmoid
            elif activation == 'relu':
                self._activation=_relu
                self._dactivation=_drelu
            else:
                raise ValueError("Unknown activation type: "+activation)
        else:
            if not _HAS_CUDAMAT:
                raise ValueError("Cudamat not available, can only run ANN with numpy")
            if activation == 'softplus':
                self._activation=cm.log_1_plus_exp
                self._dactivation=_sigmoid_cudamat
            elif activation == 'relu':
                self._activation=_relu_cudamat
                self._dactivation=_drelu_cudamat
            else:
                raise ValueError("Unknown activation type: "+activation)
        self.layers=layers
        self.W=[]
        self.b=[]
        for i in range(len(self.layers)-1):
            self.W.append(np.zeros(shape=(self.layers[i],self.layers[i+1])))
            self.b.append(np.zeros(shape=(self.layers[i+1]))+init_b)
        self.W.append(np.zeros(shape=(self.layers[len(layers)-1],1)))
        self.b.append(np.zeros(shape=(1))+init_b)

        n=0
        for i in range(len(self.W)):
            n+=np.prod(self.W[i].shape)
        self.nparW=n

        n=0
        for i in range(len(self.b)):
            n+=np.prod(self.b[i].shape)
        self.nparB=n

        self.npar=self.nparW + self.nparB

        self.narg=self.layers[0]

        if random_weights:
            # see Deep Learning, Eq. 8.23
            for i in range(len(self.W)):
                self.W[i]+=np.random.uniform(-1.0,1.0,size=self.W[i].shape)*np.sqrt(6/(np.sum(self.W[i].shape)))
        if self.cuda:
            self.cuda_setup()

    def cuda_setup(self):
         self.cu_W=[]
         for i in range(len(self.W)):
             self.cu_W.append(cm.CUDAMatrix(self.W[i]))
         self.cu_b=[]
         for i in range(len(self.b)):
             self.cu_b.append(cm.CUDAMatrix(np.reshape(self.b[i],(1,-1))))
    
    # set array of parameters
    def setpar(self,par):
        assert len(par)==self.npar
        n=0
        for i in range(len(self.W)):
            self.W[i]=np.reshape(par[n:n+np.prod(self.W[i].shape)],self.W[i].shape)
            n+=np.prod(self.W[i].shape)
        for i in range(len(self.b)):
            self.b[i]=np.reshape(par[n:n+np.prod(self.b[i].shape)],self.b[i].shape)
            n+=np.prod(self.b[i].shape)
        if self.cuda:
            self.cuda_setup()
        return self

    def getpar(self):
        par=np.zeros(self.npar)
        n=0
        for i in range(len(self.W)):
            m=np.prod(self.W[i].shape)
            par[n:n+m]=self.W[i].flatten()
            n+=m
        for i in range(len(self.b)):
            m=np.prod(self.b[i].shape)
            par[n:n+m]=self.b[i]
            n+=m
        return par

    # compute function and derivative wrt flatten parameters for a vector of samples
    def derpar(self,x):
        x = ensure_np_array(x)

        if len(x.shape)==1:
          f, der = self.derpar(x.reshape((-1,len(x))))
          return f[0], der[0]
        elif len(x.shape)>2:
          raise TypeError("Incorrectly shaped x")

        assert x.shape[1]==self.narg
        f,df_dW,df_db=self.deriv(x)
        der=np.zeros((x.shape[0],self.npar))
        n=0
        for i in range(len(self.W)):
            m=np.prod(self.W[i].shape)
            der[:,n:n+m]=df_dW[i].reshape(x.shape[0],-1)
            n+=m
        for i in range(len(self.b)):
            m=np.prod(self.b[i].shape)
            der[:,n:n+m]=df_db[i].reshape(x.shape[0],-1)
            n+=m
        return f,der

    # evaluate the NN on a single point or on an array of points
    def apply(self,x):
        x = ensure_np_array(x)
        if len(x.shape)==1:
            return self.apply(x.reshape((1,len(x))))[0]
        elif len(x.shape)>2:
            raise TypeError("Incorrectly shaped x")
        assert x.shape[1]==self.narg

        if not self.cuda:
            for i in range(len(self.W)):
                x=np.matmul(x,self.W[i])+self.b[i]
                if i+1<len(self.W):
                    x = self._activation(x)
        else:
            cu_x=cm.CUDAMatrix(x)
            for i in range(len(self.cu_W)):
                cu_x=cm.dot(cu_x,self.cu_W[i])
                cu_x.add_row_vec(self.cu_b[i])
                if i+1<len(self.cu_W):
                    self._activation(cu_x)
            x=np.array(cu_x.asarray(),dtype=np.float)
         
        return x[:,0]

    # compute derivatives with respect to parameters
    def deriv(self,x):

        x = ensure_np_array(x)

        if len(x.shape)==1:
            f,df_dW,df_db = self.deriv(x.reshape((1,len(x))))
            for i in range(len(df_dW)):
                df_dW[i]=df_dW[i][0]
                df_db[i]=df_db[i][0]
            return f[0], df_dW, df_db
        elif len(x.shape)>2:
            raise TypeError("Incorrectly shaped x")

        assert x.shape[1]==self.narg

        vec=x.shape[0]

        # allocate hidden nodes
        h=[None]*len(self.layers) # hidden nodes
        ht=[None]*len(self.layers) # non-linear functions of hidden nodes

        # allocate derivatives
        df_dW=[None]*len(self.layers)
        df_db=[None]*len(self.layers)

        if not self.cuda:

            # forward propagation
            ht[0]=+x

            for i in range(len(self.layers)-1):
                h[i+1]=np.matmul(ht[i],self.W[i])+self.b[i]
                ht[i+1] = self._activation(h[i+1])

            f=(np.matmul(ht[-1],self.W[-1])+self.b[-1])[:,0]

            # backward propagation

            df_db[-1]=np.ones((vec,1))
            df_dW[-1]=ht[-1][:,:,np.newaxis]

            for i in reversed(range(len(self.layers)-1)):
                df_db[i]=np.matmul(df_db[i+1],self.W[i+1].T) * self._dactivation(h[i+1])
                df_dW[i]=ht[i][:,:,np.newaxis]*df_db[i][:,np.newaxis,:]

        else:

            # forward propagation
            ht[0]=cm.CUDAMatrix(x)

            for i in range(len(self.cu_W)-1):
                h[i+1]=cm.dot(ht[i],self.cu_W[i])
                h[i+1].add_row_vec(self.cu_b[i])
                ht[i+1]=h[i+1].copy()
                self._activation(ht[i+1])

            f=cm.dot(ht[-1],self.cu_W[-1]).asarray()[:,0]+self.b[-1][0]

            # backward propagation

            for i in range(len(self.cu_W)):
                ht[i]=ht[i].asarray()

            df_db_host=[None]*len(self.layers)

            df_db[-1]=cm.CUDAMatrix(np.ones((vec,1)))
            df_dW[-1]=ht[-1][:,:,np.newaxis]

            df_db_host[-1]=df_db[-1].asarray()

            for i in reversed(range(len(self.layers)-1)):
                self._dactivation(h[i+1])
                df_db[i]=cm.dot(df_db[i+1],self.cu_W[i+1].transpose())
                df_db[i].mult(h[i+1])
                df_db_host[i]=df_db[i].asarray()
 
                # this is still on CPU:
                df_dW[i]=ht[i][:,:,np.newaxis]*df_db_host[i][:,np.newaxis,:]

            df_db=df_db_host

            for i in range(len(df_dW)):
                df_dW[i]=np.array(df_dW[i],dtype=np.float)
            for i in range(len(df_db)):
                df_db[i]=np.array(df_db[i],dtype=np.float)
            f=np.array(f,dtype=np.float)
            

        return f,df_dW,df_db

    # these are aliases for backward compatibility
    def applyVec(self,x):
        return self.apply(x)
    def derivVec(self,x):
        return self.deriv(x)
    def derparVec(self,x):
        return self.derpar(x)



import os
import time
import sys
import datetime
import subprocess
from typing import Optional
import yaml
from . import coretools

def _time_to_next_event(period: int):
    now=time.localtime()
    seconds=now.tm_hour*3600 + now.tm_min*60 + now.tm_sec
    seconds = seconds%period
    return period-seconds

def _now():
    return '{}'.format(datetime.datetime.now()) + ":"

def _run(cron_file: str,
         period: int):
    try:
        print(_now(),"Running now")
        if len(cron_file) > 0:
            with open(cron_file) as rc:
                config=yaml.load(rc,Loader=yaml.BaseLoader)
        else:
            config=coretools.config()
        if "cron" in config:
           for i in range(len(config["cron"])):
              args = [sys.executable, "-c", config["cron"][i]]
              timeout=_time_to_next_event(period)//2+1
              print(_now(),"cmd " + str(i) +" with timeout " + str(timeout))
              subprocess.run(args, timeout=timeout)
    except Exception as e:
        print(e)
        print()
        print(_now(),"Wait for the next scheduled event")

def cron(*,
         quick_start: bool = False,
         cron_file: str = "",
         screen_cmd: str = "screen",
         no_screen: bool = True,
         sockname: str = "cron",
         python_exec: str = "",
         detach: bool = False,
         period: int = 3600,
         max_times: Optional[int] = None
         ):
    if no_screen:
        print(_now(),"start")
        counter=0
        if quick_start:
            _run(cron_file,period)
            counter += 1
            if max_times is not None:
                if counter >= max_times:
                    return
        while True:
            s=_time_to_next_event(period)
            print(_now(),"Waiting " +str(s)+ " seconds for next scheduled event")
            time.sleep(s)
            _run(cron_file,period)
            counter += 1
            if max_times is not None:
                if counter >= max_times:
                    return
    else:
        if python_exec == "":
            python_exec = sys.executable
        print("python_exec:", python_exec)
        cmd = screen_cmd
        if detach:
            cmd += " -d"
        cmd += " -m -S " + sockname
        cmd +=" "+ python_exec + " -m bussilab cron --no-screen"
        cmd +=" --period " + str(period)
        if len(cron_file)>0:
            cmd += " --cron-file " + cron_file
        if quick_start:
            cmd += " --quick-start"
        if max_times is not None:
            cmd += " --max-times " + str(max_times)
        print(cmd)
        os.system(cmd)



"""
Module containing a WHAM implementation.

See `bussilab.wham.wham()`.

"""
import sys
from typing import Optional
import numpy as np
from . import coretools

class WhamResult(coretools.Result):
    """Result of a `bussilab.wham.wham` calculation."""
    def __init__(self,
                 *,
                 logW: np.ndarray,
                 logZ: np.ndarray,
                 nit: int,
                 eps: float):
        super().__init__()
        self.logW = logW
        """`numpy.ndarray` containing the logarithm of the weight of the frames."""
        self.logZ = logZ
        """`numpy.ndarray` containing the logarithm of the partition function of each state."""
        self.nit = nit
        """The number of performed iterations."""
        self.eps = eps
        """The final error in the iterative solution."""

def wham(bias,
        *,
        frame_weight=None,
        traj_weight=None,
        T: float = 1.0,
        maxiter: int = 1000,
        threshold: float = 1e-40,
        verbose: bool = False,
        logZ: Optional[np.ndarray] = None,
        logW: Optional[np.ndarray] = None):
    """Compute weights according to binless WHAM.

       The main input for this calculation is in the 2D array `bias`.
       Element `bias[i, j]` should contain the energy of the i-th frame computed
       according to the j-th Hamiltonian. Trajectories should be concatenated first,
       so that the total number of frames should be equal to the number of frames
       of each trajectory multiplied by the number of trajectories.
       However, it is also possible to contatenate simulations of different lengths.
       It is crucial however to compute the potential according to each of the employed
       Hamiltonian on **all the frames**, not only on those simulated using that Hamiltonian.

       Combining trajectories of different length
       ------------------------------------------

       Let's imagine three frames obtained from three Hamiltonians. Let's assume that the
       energy of frame i in Hamiltonian j is given by `bias[i, j]` defined as
       ```python
       import numpy as np
       bias = np.array([[1, 10, 7],
                       [2, 9, 6],
                       [3, 8, 5]])
       ```
       We can compute the weights with the following command:
       ```python
       np.exp(wham.wham(bias).logW)

       array([0.41728571, 0.39684866, 0.18586563])
       ```
       We now notice that the second and third columns of this matrix are equal except for
       a rigid shift.  They thus correspond to Hamiltonians that are equivalent.
       We should have been able to obtain the same result saying that these frames were
       coming from two simulations. If we only pass the first two columns however
       we obtain different weights
       ```python
       np.exp(wham.wham(bias[:, 0:2]).logW)

       array([0.28224026, 0.43551948, 0.28224026])
       ```
       In order to correcly analyze these frames we should pass the information that
       the second Hamiltonian was used for twice the time:
       ```python
       np.exp(wham.wham(bias[:, 0:2], traj_weight=(1, 2)).logW)

       array([0.41728571, 0.39684866, 0.18586563])
       ```
       Notice that now the weights are identical to those computed in the first example.

       Trusting more a trajectory than another
       ---------------------------------------

       When you concatenate trajectories, you might explicitly want to trust more a trajectory
       than another.  For instance, two trajectories might have been accumulated with a different
       stride, and the reliability of each frame of the one with smaller stride should be lower.
       We thus would like to assign a weight to each frame that accounts for its reliability.

       Consider the following command:
       ```python
       import numpy as np
       np.exp(wham.wham([[3, 5],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4],
                         [4, 4]]).logW)
       ```
       that results in the following weights
       ```python
       array([0.06421006, 0.09357899, 0.09357899, 0.09357899, 0.09357899,
              0.09357899, 0.09357899, 0.09357899, 0.09357899, 0.09357899,
              0.09357899])
       ```
       Notice that all the frames except for the first one are identical. An equivalent result
       would have been obtained using
       ```python
       import numpy as np
       np.exp(wham.wham([[3, 5],
                         [4, 4]], frame_weight=(1, 10)).logW)

       array([0.06421006, 0.93578994])
       ```
       Clearly, the weight of the second frame in this example is equal to ten times the
       weights of the ten corresponding frames in the previous example.


       Parameters
       ----------

       bias: np.ndarray
           An array with shape (nframes, ntraj) containing the bias potential applied to
           each frame according to each of the Hamiltonians.

       frame_weight: np.ndarray, optional
           An array with nframes elements. These elements should contain the reliability weight
           of the frames.  By default, these weights are set to one.

       traj_weight: np.ndarray, optional
           An array with ntraj elements. These elements should contain the total weight of each of
           the Hamiltonians. Should be used when combining trajectories of different lengths.

       T: float, optional
           The temperature of the system. This number is just used to divide the `bias` array in
           order to make it adimensional.

       maxiter: int, optional
           Maximum number of iterations in the minimization procedure.

       threshold: float, optional
           Threshold for the minimization procedure.

       verbose: bool, optional
           If True, print information as the minimization proceeds.

       logZ: np.ndarray, optional
           Array with ntraj elements.
           Initial value for the logarithm of the partition functions. If not provided, it is
           computed from the bias.  Providing an initial guess that is close to the converged
           value (e.g. as obtained from a calculation with a limited number of frames) can speed up
           significantly the convergence.

       logW: np.ndarray, optional
           Array with nframes elements.
           Initial value for the logarithm of the weights. If not provided, they are computed from
           the bias.  Providing an initial guess that is close to the converged value can speed up
           significantly the convergence. *If logW is provided, logZ is ignored*.

    """

    # allow tuples or lists
    bias = coretools.ensure_np_array(bias)
    frame_weight = coretools.ensure_np_array(frame_weight)
    traj_weight = coretools.ensure_np_array(traj_weight)

    nframes = bias.shape[0]
    ntraj = bias.shape[1]

    # default values
    if frame_weight is None:
        frame_weight = np.ones(nframes)
    if traj_weight is None:
        traj_weight = np.ones(ntraj)

    assert len(traj_weight) == ntraj
    assert len(frame_weight) == nframes

    # do exponentials only once
    expv = np.exp((-bias+np.min(bias))/T)

    if logW is not None:
        Z = np.matmul(np.exp(logW), expv)
        Z /= np.sum(Z*traj_weight)
    elif logZ is not None:
        Z = np.exp(logZ)
    else:
        Z = np.ones(ntraj)

    Zold = Z

    if verbose:
        sys.stderr.write("WHAM: start\n")
    for nit in range(maxiter):
        # find unnormalized weights
        weight = 1.0/np.matmul(expv, traj_weight/Z)*frame_weight
        # normalized them
        weight /= np.sum(weight)
        # update partition functions
        Z = np.matmul(weight, expv)
        # normalize the partition functions
        Z /= np.sum(Z*traj_weight)
        # monitor change in partition functions
        eps = np.sum(np.log(Z/Zold)**2)
        Zold = Z
        if verbose:
            sys.stderr.write("WHAM: iteration "+str(nit)+" eps "+str(eps)+"\n")
        if eps < threshold:
            break
    if verbose:
        sys.stderr.write("WHAM: end")

    return WhamResult(logW=np.log(weight), logZ=np.log(Z), nit=nit, eps=eps)



"""
Module containing a tool to solve Potts models by enumeration

See `bussilab.potts.Model()`.

"""

import numpy as np
from scipy.optimize import minimize
from typing import Optional, Callable
import numba
import warnings

from . import coretools

@numba.jit
def _make_lists(size: int,
                colors: int = 1,
                start: int = 0,
                stop: int = -1,
                shifted: bool = False):
    N=(colors+1)**size
    if(stop==-1):
        stop=N
    ret=np.zeros((stop-start,size*colors))
    for i in range(stop-start):
        l=i+start
        for j in range(size):
            m=l%(colors+1)
            if m>0:
                ret[i,(size-j-1)*colors+m-1]=1.0
            l=l//(colors+1)
    if shifted:
        ret-=0.5
    return ret

class InferResult(coretools.Result):
    """Result of a `bussilab.potts.Model.infer` calculation."""
    def __init__(self,
                 *,
                 h: np.ndarray,
                 J: np.ndarray,
                 averages: np.ndarray,
                 loglike: float,
                 success: bool,
                 message: str,
                 nfev: int,
                 nit: int):
        super().__init__()
        self.h = h
        """`np.ndarray`, optimized h."""
        self.J = J
        """`np.ndarray`, optimized h."""
        self.averages = averages
        """`np.ndarray`, resulting averages."""
        self.loglike = loglike
        """`float` containing the resulting likelihood Gamma."""
        self.success = success
        """`bool` reporting the success of the minimizer."""
        self.message = message
        """`str` reporting the possible reason of failuer of the minimizer."""
        self.nfev = nfev
        """`int` reporting the number of function evaluations."""
        self.nit = nit
        """`int` reporting the number of iterations in the minimization procedure."""

class Model:
    def __init__(self,
                 size: int,
                 colors: int = 1,
                 shifted: bool = False,
                 fullmatrix: bool = True):
        """Init model.
           size: number of spins
           colors: number of colors
           fullmatrix: set to False to use less memory (slower)
        """
        if colors != 1:
            warnings.warn("number of colors different from 1 is not tested")
        if shifted:
            warnings.warn("shifted spins not tested")
        self.size=size
        self.colors=colors
        self.nstates=(1+colors)**self.size
        # list of all possible sequences
        self.allseq=_make_lists(self.size,colors,shifted=shifted)
        # outer products of all possible sequences
        # used to compute the energy
        # takes a lot of memory but allow to write operations with numpy
        if fullmatrix:
            self.allseq_matrix=np.einsum('ki,kj->kij',self.allseq,self.allseq)
        else:
            self.allseq_matrix=None
    def compute(self,
                h: np.ndarray,
                J: np.ndarray):
        """Compute averages <sigma_i,sigma_j> for a coupling matrix J.
           Returns (a,b) with a=free energy and b=averages
        """
        if not self.allseq_matrix is None:
            all_ene=np.tensordot(self.allseq_matrix,self.fixJ(J),((1,2),(0,1)))
        else:
            all_ene=np.einsum("ki,kj,ij->k",self.allseq,self.allseq,self.fixJ(J))
        if h is not None:
            all_ene+=np.matmul(self.allseq,h.T)
        shift=all_ene.min()
        all_ene-=shift
        prob=np.exp(-all_ene)
        Z=np.sum(prob)
        if not self.allseq_matrix is None:
            average=np.tensordot(prob,self.allseq_matrix,(0,0))/Z
        else:
            average=np.einsum("ki,kj,k->ij",self.allseq,self.allseq,prob)/Z
        return (-np.log(Z)+shift,average)
    def loglike(self,
                h: np.ndarray,
                J: np.ndarray,
                ave: np.ndarray):
        """Compute -log likelihood for a coupling matrix J with averages ave.
           Returns (a,b) with a=-log likelihood and b=derivatives
        """
        c=self.compute(h,self.fixJ(J))
        l=np.sum(self.fixJ(J)*ave)
        if h is not None:
            l+=np.sum(h*np.diag(ave))
        der=-c[1]+ave
        return (l-c[0],der)
    def draw(self,
             h: np.ndarray,
             J: np.ndarray,
             n: int):
        """Compute averages for a coupling matrix J sampling n states.
        """
        if not self.allseq_matrix is None:
            all_ene=np.tensordot(self.allseq_matrix,self.fixJ(J),((1,2),(0,1)))
        else:
            all_ene=np.einsum("ki,kj,ij->k",self.allseq,self.allseq,self.fixJ(J))
        if h is not None:
            all_ene+=np.matmul(self.allseq,h.T)
        shift=all_ene.min()
        all_ene-=shift
        prob=np.exp(-all_ene)
        Z=np.sum(prob)
        prob/=Z
        ret=np.zeros((self.size*self.colors,self.size*self.colors))
        for i in range(n):
            j=np.random.choice(len(self.allseq),p=prob)
            if not self.allseq_matrix is None:
                ret+=self.allseq_matrix[j]
            else:
                ret+=np.outer(self.allseq[j],self.allseq[j])
        return ret/n
    def fixJ(self,
             J: np.ndarray):
        newJ=0.5*(J+J.T)
        for i in range(self.size):
            newJ[self.colors*i:self.colors*(i+1),self.colors*i:self.colors*(i+1)]=0.0
        return newJ
    def infer(self,
              averages: np.ndarray,
              nseq: int = 1,
              reg: Optional[Callable] = None):
        x0=np.zeros(self.size*self.size*self.colors*self.colors)
        def function(par,m,a):
            J=m.fixJ(par.reshape((self.size*self.colors,self.size*self.colors)))
            h=np.diag(par.reshape((self.size*self.colors,self.size*self.colors)))
            c=m.loglike(h,J,a)
            c=(nseq*c[0],nseq*c[1])
            if reg is not None:
                r=reg(J)
                c=(c[0]+r[0],c[1]+r[1])
            return (c[0],c[1].flatten())
        res = minimize(function, x0, args=(self,averages),
              method="L-BFGS-B",jac=True,tol=1e-10)
        h=np.diag(res.x.reshape((self.size*self.colors,self.size*self.colors)))
        J=self.fixJ(res.x.reshape((self.size*self.colors,self.size*self.colors)))
        return InferResult(
            h=h,
            J=J,
            averages=averages,
            loglike=res.fun,
            success=res.success,
            message=res.message,
            nfev=res.nfev,
            nit=res.nit
            )
    def random_couplings(self,
                         seed: Optional[int] = None):
        if seed is not None:
            np.random.seed(seed)
        J=np.triu(np.random.normal(0,1.0,(self.size*self.colors,self.size*self.colors)))
        return self.fixJ(J)



"""
Module implementing tools for remote jupyter connections.

This module contains some utilities that are mostly designed
to be used as command line tools. The interface of the functions defined in
this module is subject to changes. One should instead use the subcommands
`jrun` and `jremote` in the [cli_documentation](cli_documentation.html)
submodule.
"""
import socket
import os
from contextlib import closing
import sys
import subprocess
import re
import platform
import time

def find_free_port():
    """Returns the number of a free port."""
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]

def run_server(dry_run: bool = False,
               port: int = 0,
               screen_cmd: str = "screen",
               no_screen: bool = False,
               python_exec: str = "",
               sockname: str = "jupyter-server",
               detach: bool = False):
    """Runs a jupyter server inside a screen command.

       This function is only designed to be used as a command line
       tool.
    """
    if python_exec == "":
        python_exec = sys.executable
    print("python_exec:", python_exec)
    if port == 0:
        port = find_free_port()
    print("port:", port)
    cmd = ""
    if not no_screen:
        cmd += screen_cmd
        if detach:
            cmd += " -d"
        cmd += " -m -S " + sockname + "-" + str(port) + " "
    cmd += python_exec + " -m jupyter notebook --no-browser "
    cmd += "--port=" + str(port)
    print("cmd:", cmd)
    if dry_run:
        return
    os.system(cmd)

def remote(server: str,
           python_exec: str = "python",
           dry_run: bool = False,
           list_only: bool = False,
           port: int = 0,
           index: int = 0,
           open_cmd: str = ""):

    cmd = python_exec+" -m jupyter notebook list"

    print("server:", server)
    print("cmd:", cmd)

    if dry_run and list_only:
        return

    ll = []
    ll_localhost = []
    args = ['ssh', server, cmd]

    for l in subprocess.run(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            universal_newlines=True,
                            check=True).stdout.split('\n'):
        if re.match("^http", l):
            ll.append(re.sub("localhost", server, l))
            ll_localhost.append(l)

    for i in range(len(ll)):
        print(str(i+1)+") "+ll[i])

    if list_only:
        return

    if len(ll_localhost) == 1:
        index = 1

    if index == 0:
        sys.stdout.write("Choose a notebook or interrupt (^c):")
        sys.stdout.flush()
        index = int(sys.stdin.readline())

    url = ll_localhost[index-1].split()[0]

    print("Chosen url:", url)

    if not re.match(r"^http://localhost:[0-9]*/\?token=[0-9a-f]*$", url):
        raise Exception("URL "+url+" looks incorrectly formatted")

    server_port = re.sub("^http://localhost:", "", url)
    server_port = re.sub("/.*$", "", server_port)

    if port == 0:
        port = find_free_port()

    ssh_cmd = ["ssh", "-NL", str(port) + ":localhost:" + str(server_port), server]

    if open_cmd == "":
        plat = platform.system()
        if plat == "Darwin":
            open_cmd = "open"
        elif plat == "Linux":
            open_cmd = "xdg-open"
        else:
            raise Exception("Unknown platform " + plat)
    print("open_cmd:", open_cmd)

    open_cmd = open_cmd + " " + re.sub(":" + str(server_port), ":" + str(port), url)
    print("ssh:", ' '.join(ssh_cmd))
    if dry_run:
        print(open_cmd)
    else:
        ssh = subprocess.Popen(ssh_cmd)
        try:
            print("Don't kill this terminal or you will be disconnected!")
            time.sleep(1)
            os.system(open_cmd)
            print("Use CTRL+c to kill the connection")
        except KeyboardInterrupt:
            print("Interrupted")
        finally:
            try:
                ssh.communicate()
            except KeyboardInterrupt:
                print("Interrupted")



"""
Tools to perform reweighting using MaxEnt.
"""
import sys
from typing import Optional, Callable
import numpy as np
from scipy.optimize import minimize
from . import coretools

class MaxentResult(coretools.Result):
    """Result of a `bussilab.maxent.maxent` calculation."""
    def __init__(self,
                 *,
                 logW_ME: np.ndarray,
                 lambdas: np.ndarray,
                 averages: np.ndarray,
                 gamma: float,
                 success: bool,
                 message: str,
                 nfev: int,
                 nit: int):
        super().__init__()
        self.logW_ME = logW_ME
        """`np.ndarray` with `traj.shape[0]` elements, logarithms of the optimized weights."""
        self.lambdas = lambdas
        """`np.ndarray` with `len(reference)` elements, optimized Lagrangian multipliers."""
        self.averages = averages
        """`np.ndarray` with `len(reference)` elements, resulting averages."""
        self.gamma = gamma
        """`float` containing the resulting likelihood Gamma."""
        self.success = success
        """`bool` reporting the success of the minimizer."""
        self.message = message
        """`str` reporting the possible reason of failuer of the minimizer."""
        self.nfev = nfev
        """`int` reporting the number of function evaluations."""
        self.nit = nit
        """`int` reporting the number of iterations in the minimization procedure."""


# Internal tool to compute averages over trajectory.
# Does not access external data.
# Might be optimized on GPU or to access traj from disk.
def _heavy_part(logW: np.ndarray,
                traj: np.ndarray,
                l: np.ndarray,
                weights: bool = False):
    logW_ME = logW-np.dot(traj, l) # maxent correction
    shift_ME = np.max(logW_ME) # shift to avoid overflow
    W_ME = np.exp(logW_ME - shift_ME)
    # Partition function:
    Z = np.sum(W_ME)
    # Averages:
    averages = np.dot(W_ME, traj) / Z
    logZ = np.log(Z) + shift_ME
    if not weights:
        return (logZ, averages)
    # only return weights if requested:
    logW_ME -= np.log(Z)+shift_ME
    return (logZ, averages, logW_ME)

def maxent(
        traj,
        reference,
        *,
        logW=None,
        maxiter: int = 1000,
        verbose: bool =False,
        lambdas=None,
        l2=None,
        l1=None,
        method: str = "L-BFGS-B",
        regularization: Optional[Callable] = None,
        tol: Optional[float] = None,
        options=None):
    """Tool that computes new weights to enforce reference values.

       This tools process a an array containing the observables computed along a trajectory and
       returns new weights that satisfy the maximum entropy principle and so that weighted averages
       agree with reference values.

       Parameters
       ----------

       traj : array_like
           A 2D array (lists or tuples are internally converted to numpy arrays).
           `traj[i,j]` is j-th observable computed in the i-th frame.

       reference : array_like

           A 1D array (lists or tuples are internally converted to numpy arrays)
           containing the reference values to be enforced. If the i-th element is a tuple
           or an array with 2 elements, they are interpreted as boundaries. For instance,
           `reference=[1.0,(2.0,3.0)]` will make sure the first observable has value 1 and
           the second observable is within the range (2,3). Boundaries equal to `+np.inf`
           or `-np.inf` can be used to imply no boundary. Notice that boundaries in the
           form (A,B) where both A and B are finite are implemented by adding fictitious
           variables in a way that is transparent to the user.  Boundaries in the form
           (A,B) where one of A or B is finite and the other is infinite are implemented
           as boundaries on lambdas.  Boundaries in the form (A,A) are interpreted as
           constraints.

       logW : array_like

           A 1D array (lists or tuples are internally converted to numpy arrays)
           containing the logarithm of the a priori weight of the provided frames.

       lamdbas : array_like

           A 1D array with initial values of lambda. A good guess will minimize faster. A
           typical case would be recycling the lambdas obtained with slighlty different
           regularization parameters.

       l2 : None, float, or array_like

           Prefactor for L2 regularization. If None, no regularization is applied. If
           float, the same factor is used on all the lambdas.  If it is an array, it
           should have length equal to `len(reference)`.

       l1 : None, float, or array_like

           Prefactor for L1 regularization. If None, no regularization is applied. If
           float, the same factor is used on all the lambdas.  If it is an array, it
           should have length equal to `len(reference)`.

       regularization : callable or None

           A function that takes as argument the current lambdas and return an tuple
           containing the regularization function and its derivatives. For instance,
           passing a function defined as `def reg(x): return
           (1e-4*0.5*np.sum(x**2),1e-4*x)` is equivalent to passing `l2=1e-4`.

       verbose : bool
           If True, progress informations are written on stdout.

       method : str
           Minimization method. See documentation of `scipy.optimize.minimize`.

       maxiter : int
           Maximum number of iterations

       tol : float or None
           Tolerance for minimization. See documentation of scipy.optimize.minimize.

       options : dict
           Arbitrary options passed to `scipy.optimize.minimize`.

    """

    traj = coretools.ensure_np_array(traj)
    lambdas = coretools.ensure_np_array(lambdas)

    nframes = traj.shape[0]
    nobs = traj.shape[1]

    # accepts a scalar as l2 regularization term
    if isinstance(l2, float):
        l2 = np.ones(nobs)*l2

    l2 = coretools.ensure_np_array(l2)

    if isinstance(l1, float):
        l1 = np.ones(nobs)*l1

    l1 = coretools.ensure_np_array(l1)
    # default values
    if logW is None:
        logW = np.zeros(nframes)
    if lambdas is None:
        lambdas = np.zeros(nobs)

    # checks
    assert len(reference) == nobs
    assert len(logW) == nframes
    assert len(lambdas) == nobs

    fullreference = []
    bounds = []
    box_const = []
    for i in range(nobs):
        if hasattr(reference[i], "__len__"):
            if len(reference[i]) > 1:
                if len(reference[i]) > 2:
                    raise TypeError("")
                if reference[i][0] > reference[i][1]:
                    raise TypeError("")
                if reference[i][0] == reference[i][1]:
                    fullreference.append(reference[i][0])
                    bounds.append((-np.inf, +np.inf))
                    box_const.append(False)
                elif (np.isinf(reference[i][1]) and reference[i][1] > 0.0
                      and not np.isinf(reference[i][0])):
                    fullreference.append(reference[i][0])
                    bounds.append((-np.inf, 0.0))
                    box_const.append(False)
                elif (np.isinf(reference[i][0]) and reference[i][0] < 0.0
                      and not np.isinf(reference[i][1])):
                    fullreference.append(reference[i][1])
                    bounds.append((0.0, +np.inf))
                    box_const.append(False)
                elif not np.isinf(reference[i][0]) and not np.isinf(reference[i][1]):
                    fullreference.append(reference[i][0])
                    fullreference.append(reference[i][1])
                    bounds.append((-np.inf, 0.0))
                    bounds.append((0.0, +np.inf))
                    box_const.append(True)
                elif ((np.isinf(reference[i][0]) and reference[i][0] < 0.0)
                      and (np.isinf(reference[i][1]) and reference[i][1] > 0.0)):
                    fullreference.append(0.0)
                    bounds.append((0.0, 0.0))
                    box_const.append(False)
                else:
                    raise TypeError("")
            else:
                fullreference.append(reference[i][0])
                bounds.append((-np.inf, +np.inf))
                box_const.append(False)
        else:
            fullreference.append(reference[i])
            bounds.append((-np.inf, +np.inf))
            box_const.append(False)


    fullreference = np.array(fullreference)
    bounds = np.array(bounds)
    box_const = np.array(box_const)

    nit = 0
    def _callback(par):
        nonlocal nit # needed to access outer scope
        nit += 1
        if verbose:
            sys.stderr.write("MAXENT: iteration "+str(nit)+"\n")

    callback: Optional[Callable] = None
    # verbose logging
    if verbose:
        sys.stderr.write("MAXENT: start\n")
        callback = _callback

    # logZ0 is not changing during minimization and is computed once.
    # it is only needed to compute Gamma
    shift0 = np.max(logW) # shift to avoid overflow
    W0 = np.exp(logW - shift0)
    logZ0 = np.log(np.sum(W0)) + shift0

    # function to be minimized
    def func(l):

        l = np.array(l) # ensure array

        assert len(l) == len(fullreference)

        # takes care of >< constraints
        # vector ll only contains the Lagrangian multipliers to be applied on the trajectory
        if len(fullreference) != nobs:
            ll = np.zeros(nobs)
            s = 0
            for i in range(nobs):
                if box_const[i]:
                    # >< multipliers are summed
                    ll[i] = l[i+s] + l[i+s+1]
                    s += 1
                else:
                    ll[i] = l[i+s]
        else:
            ll = l

        logZ, averages = _heavy_part(logW, traj, ll)

        f = logZ - logZ0
        der = -averages

        if regularization is not None:
            reg = regularization(ll)
            f += reg[0]
            der += reg[1]

        if l2 is not None:
            f += 0.5*np.sum(l2*ll**2)
            der += l2*ll

        if l1 is not None:
            eee = 1e-50
            f += np.sum(l1*np.sqrt(ll**2+eee**2))
            der += l1*ll/np.sqrt(ll**2+eee**2)

        # takes care of >< constraints
        # vector der only contains the nobs elements
        # it is here extended
        if len(fullreference) != nobs:
            newder = np.zeros(len(fullreference))
            s = 0
            for i in range(nobs):
                if box_const[i]:
                    newder[i+s] = der[i]
                    newder[i+s+1] = der[i]
                    s += 1
                else:
                    newder[i+s] = der[i]
            der = newder

        # fullreference contains already nobs+nshift elements
        f += np.dot(l, fullreference)
        der += fullreference

        return(f, der)


    # With >< constraints the initial lambdas should be fixed
    if len(fullreference) != nobs:
        ll = np.zeros(len(fullreference))
        s = 0
        for i in range(nobs):
            if box_const[i]:
                if lambdas[i] >= 0:
                    ll[i+s+1] = lambdas[i]
                else:
                    ll[i+s] = lambdas[i]
                s += 1
            else:
                ll[i+s] = lambdas[i]
        lambdas = ll

    if maxiter is not None:
        if options is None:
            options = {}
        options["maxiter"] = maxiter

    res = minimize(
        func, lambdas, method=method, jac=True, callback=callback, bounds=bounds,
        tol=tol, options=options)

    # With >< constraints the final lambdas should be fixed
    if len(fullreference) != nobs:
        lambdas = np.zeros(nobs)
        s = 0
        for i in range(nobs):
            if box_const[i]:
                lambdas[i] = res.x[i+s]+res.x[i+s+1]
                s += 1
            else:
                lambdas[i] = res.x[i+s]
    else:
        lambdas = res.x

    # recompute weights at the end
    logZ, averages, logW_ME = _heavy_part(logW, traj, lambdas, weights=True)

    if verbose:
        sys.stderr.write("MAXENT: end")

    return MaxentResult(
        logW_ME=logW_ME,
        lambdas=lambdas,
        averages=averages,
        gamma=res.fun,
        success=res.success,
        message=res.message,
        nfev=res.nfev,
        nit=res.nit
    )



"""
General purpose tools.
"""
from contextlib import contextmanager
import gzip
import os
import unittest
import pathlib
import re
from typing import List, Optional

import numpy as np
import yaml

def ensure_np_array(arg) -> np.ndarray:
    """Convert arg to np.array if necessary."""
    if arg is not None and not isinstance(arg, (np.ndarray, np.generic)):
        return np.array(arg)
    return arg

def file_or_path(arg, mode: str):
    """Convert a path to an open file object if necessary."""
    if isinstance(arg, str):
        arg = open(arg, mode)
    if isinstance(arg, bytes):
        arg = open(str(arg), mode)
    if re.match(r".*\.gz", arg.name):
        arg = gzip.open(arg, mode)
    return arg

class Result(dict):
    # triple ' instead of triple " to allow using docstrings in the example
    '''Base class for objects returning results.

       It allows one to create a return type that is similar to those
       created by `scipy.optimize.minimize`.
       The string representation of such an object contains a list
       of attributes and values and is easy to visualize on notebooks.

       Examples
       --------

       The simplest usage is this one:

       ```python
       from bussilab import coretools

       class MytoolResult(coretools.Result):
           """Result of a mytool calculation."""
           pass

       def mytool():
           a = 3
           b = "ciao"
           return MytoolResult(a=a, b=b)

       m=mytool()
       print(m)
       ```

       Notice that the class variables are dynamic: any keyword argument
       provided in the class constructor will be processed.
       If you want to enforce the class attributes you should add an explicit
       constructor. This will also allow you to add pdoc docstrings.
       The recommended usage is thus:

       ````
       from bussilab import coretools

       class MytoolResult(coretools.Result):
           """Result of a mytool calculation."""
           def __init__(a, b):
               super().__init__()
               self.a = a
               """Documentation for attribute a."""
               self.b = b
               """Documentation for attribute b."""

       def mytool():
           a = 3
           b = "ciao"
           return MytoolResult(a=a, b=b)

       m = mytool()
       print(m)
       ````

    '''

    def __getattr__(self, name: str):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, item: str, value):
        self[item] = value

    def __delattr__(self, item: str):
        del self[item]

    def __repr__(self) -> str:
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
# when used recursively, the inner repr is properly indented:
            return '\n'.join([k.rjust(m) + ': ' + re.sub("\n", "\n"+" "*(m+2), repr(v))
                              for k, v in sorted(self.items())])
        return self.__class__.__name__ + "()"

    def __dir__(self) -> List[str]:
        return list(sorted(self.keys()))

@contextmanager
def cd(newdir: os.PathLike):
    """Context manager to temporarily change working directory.

       Can be used to change working directory temporarily making sure that at the
       end of the context the working directory is restored. Notably,
       it also works if an exception is raised within the context.

       Parameters
       ----------

       newdir : path
           Path to the desired directory.

       Examples
       --------

       ```python
       from bussilab.coretools import cd
       with cd("/path/to/dir"):
           do_something() # this is executed in the /path/to/dir directory
       do_something_else() # this is executed in the original directory
       ```
    """
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)

class TestCase(unittest.TestCase):
    """Improved base class for test cases.

       Extends the `unittest.TestCase` class with some additional assertion.

    """
    def assertEqualFile(self, file1: os.PathLike, file2: Optional[os.PathLike] = None):
        """Check if two files are equal.

           Parameters
           ----------

           file1: path
               Path to the first file

           file2: path, optional
               Path to the second file. If not provided, defaults to `file1+".ref"`.
        """
        if file2 is None:
            file2 = pathlib.PurePath(str(file1)+".ref")
        with open(file1, "r") as f1:
            with open(file2, "r") as f2:
                self.assertEqual(f1.read(), f2.read())


def config(path: Optional[os.PathLike] = None):
    if path is None:
        path = pathlib.PurePath(os.environ["HOME"]+"/.bussilabrc")
    with open(path) as rc:
        return yaml.load(rc,Loader=yaml.BaseLoader)




"""
Module with some clustering tools
"""

from typing import Optional

import networkx
import numpy as np

from .coretools import Result

class ClusteringResult(Result):
    """Result of a `bussilab.clustering` calculation."""
    def __init__(self,
                *,
                method: str,
                clusters: list,
                weights: Optional[list]
                ):
        self.method = method
        """`str` containing the name of the method used."""
        self.clusters = clusters
        """`list of lists` containing the members of each cluster."""
        self.weights = weights
        """`list` containing the weights of the clusters."""

def max_clique(adj,weights=None):
    """Same algorithm as in Reisser, et al, NAR (2020)."""
    # weights: optional weights
    # if adj is a graph, it will be copied
    graph=networkx.Graph(adj)
    cliques=[]
    ww=[]
    while graph.number_of_nodes()>0:
        maxw=0.0
        for i in networkx.algorithms.clique.find_cliques(graph):
            if weights is not None:
                w=np.sum(weights[i])
            else:
                w=len(i)
            if w > maxw:
                maxi=i
                maxw=w
        cliques.append(maxi)
        ww.append(maxw)
        graph.remove_nodes_from(maxi)
    return ClusteringResult(method="max_clique",clusters=cliques, weights=ww)

def daura(adj,weights=None):
    """Same algorithm as in Daura et al, Angew. Chemie (1999)."""
    adj=adj.copy() # take a copy (adj is modified while clustering)
    indexes=np.arange(len(adj))
    clusters=[]
    ww=[]
    if weights is not None:
        weights = weights.copy() # take a copy (weights is modified while clustering)
    while len(indexes)>0:
        if weights is not None:
            d=np.sum(adj*weights,axis=0)
        else:
            d=np.sum(adj,axis=0)
        n=np.argmax(d)
        ww.append(d[n])
        ii=np.where(adj[n]>0)[0]
        clusters.append(indexes[ii])
        adj=np.delete(adj,ii,axis=0)
        adj=np.delete(adj,ii,axis=1)
        if weights is not None:
            weights=np.delete(weights,ii)
        indexes=np.delete(indexes,ii)
    return ClusteringResult(method="daura",clusters=clusters, weights=ww)



"""
Module implementing Lohman model for helicases.
"""

from collections.abc import Sequence
from typing import Union, Tuple, overload
import numpy as np

@overload
def lohman(t: float, ku: float, kd: float, n: int, boundaries: Tuple[float, float]) -> float:
    pass
@overload
def lohman(t: Sequence, ku: float, kd: float, n: int, boundaries: Tuple[float, float]) -> np.array:
    pass
@overload
def lohman(t: np.array, ku: float, kd: float, n: int, boundaries: Tuple[float, float]) -> np.array:
    pass

# actual implementation:
def lohman(t,
           ku: float,
           kd: float,
           n: int = 1,
           boundaries: Tuple[float, float] = (0.0, 1.0)
          ) -> Union[float, np.array]:
    """Lohman model for helicases.

       Compute the fraction of unwound helices as a function of time.
       See [Lucius et al, Biophys J 2003](https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1303449/).

       Parameters
       ----------
       t: float or sequence or np.ndarray
           Time. If a sequence or np.ndarray is provided, the function is computed
           for all values and an array is returned.

       ku: float
           Unwinding rate.

       kd: float
           Dissociation rate.

       n: int, optional
           Step size

       boundaries: tuple with 2 elements
           Result is mapped to this range.

       Returns
       -------

       float or np.ndarray
           The fraction of unfolded helices at a given time `t`. If an array is provided for `t`,
           an array is returned containing the fractions at all the times.
           If `boundaries` is provided, the fraction is linearly mapped into the
           `boundaries[0], boundaries[1]` range.

    """
    if isinstance(t, Sequence):
        t = np.array(t, dtype=float)
    kobs = ku+kd
    # processivity
    P = ku/kobs
    f = 0*t
    fact = 1
    for r in range(n):
        f += (kobs*t)**r/fact
        fact *= r+1
    f *= np.exp(-kobs*t)
    f = P**n*(1-f)
    return boundaries[0] + (boundaries[1]-boundaries[0]) * f



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

def version():
    from bussilab import __version__ as version
    return version

def deps():
    from bussilab import _required_ as required
    return required

def description():
    from bussilab import __doc__ as doc
    return doc.partition('\n')[0]

setup(
    name="bussilab",
    version=version(),
    author='Giovanni Bussi',
    author_email='bussi@sissa.it',
    description=description(),
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/bussilab/py-bussilab',
    packages=['bussilab'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering"
        ],
    install_requires=deps(),
    python_requires='>=3.6',
    scripts=['bin/bussilab']
)



"""
Documentation for command line tools
====================================

This module only contains the documentation of the subcommands used in the
command line interface.
For all the subcommands, a short help identical to that generated by the
`-h` option is shown. In addition, if present, the docstring of the
function is also shown here.

_subparsers_help_
"""

import re
import inspect
from .cli import _create_eparser

# Then we process the docstring in order to include the generated documentation
def _process_doc(doc):
    eparser = _create_eparser("bussilab")
    _subparsers_help_ = ""
    for name in eparser.subparsers_help.keys():
        _subparsers_help_ = (
            _subparsers_help_+"## "+name+"\n```text\n"+eparser.subparsers_help[name]+"```\n"
        )
        _subparsers_help_ = (
            _subparsers_help_+"\n" +
            inspect.cleandoc(eparser.subparsers_doc[name])+"\n"
        )
    doc = re.sub("_subparsers_help_", _subparsers_help_, doc)
    return doc

__doc__ = _process_doc(__doc__)



import subprocess
import re

from . import coretools

def _parse_cpu(line: str):
#  us, user    : time running un-niced user processes
#  sy, system  : time running kernel processes
#  ni, nice    : time running niced user processes
#  id, idle    : time spent in the kernel idle handler
#  wa, IO-wait : time waiting for I/O completion
#  hi : time spent servicing hardware interrupts
#  si : time spent servicing software interrupts
#  st : time stolen from this vm by the hypervisor
    field=["us", "sy", "ni", "id", "wa", "hi", "si", "st"]
    result={}
    for i in range(1,len(line)):
        for f in field:
            if re.match("^" + f + ".*",line[i]):
                result[f]=float(line[i-1])
    return result

def workstations(wks = None, short=True):
    msg=""

    if not wks:
        c=coretools.config()
        wks=c["workstations"]

    for w in wks:
       if isinstance(w, str):
          name=w
          url=w
          disk="/scratch"
          nvidia='True'
       elif isinstance(w, dict):
          name=w['name']
          url=w['url']
          try:
              disk=w['disk']
          except KeyError:
              disk="/scratch"
          try:
              nvidia=w['nvidia']
          except KeyError:
              nvidia='True'
       else:
           raise TypeError()

       cmd = "top -n 1 -b | head -n 3 | tail -n 1;"
       cmd += "df -h " + disk + " | tail -n 1;"

       if nvidia != "False":
           cmd += "nvidia-smi  | grep Default"

       args=['ssh',url, cmd]
       msg+=name
       try:
          out=subprocess.run(args,
             stdout=subprocess.PIPE,
             stderr=subprocess.PIPE,
             universal_newlines=True,
             timeout=20,
             check=True).stdout.split('\n')
          cpu_fields=_parse_cpu(out[0].split())
          if cpu_fields["id"] > 75:
              msg += " CPU :sleeping:"
          elif cpu_fields["id"] > 50:
              msg += " CPU :walking:"
          else:
              msg += " CPU :running:"
          if cpu_fields["id"] + cpu_fields["us"] + cpu_fields["ni"] < 80:
              msg += "(:warning: id+us+ni<80)"
          if nvidia != 'False':
              for j in range(2,len(out)-1):
                  gpu_fields=out[j].split()
                  gpu_usage=0.0
                  for i in range(1,len(gpu_fields)):
                      if re.match("^Default",gpu_fields[i]):
                          gpu_usage=float(gpu_fields[i-1].strip("%"))
                  if j<=2:
                      msg += " GPU"
                  if gpu_usage < 25:
                      msg += " :sleeping:"
                  elif gpu_usage < 50:
                      msg += " :walking:"
                  else:
                      msg += " :running:"
          msg += " disk"
          disk_fields=out[1].split()
          disk_occupation=float(disk_fields[-2].strip("%"))
          if disk_occupation < 80:
              msg += " :smile:"
          elif disk_occupation < 90:
              msg += " :neutral_face:"
          elif disk_occupation < 99:
              msg += " :worried:"
          else:
              msg += " :scream:"
          msg += "\n"
          if not short:
              msg+="\n".join(out)+"\n\n"
       except subprocess.TimeoutExpired:
          msg += " :skull_and_crossbones:"
          msg += "\n"
          msg += "Timeout"
          msg +="\n\n"
       except Exception:
          msg += " :skull_and_crossbones:"
          msg += "\n"
          msg +="Error connecting"
          msg += "\n\n"
    return msg

